# -*- coding: utf-8 -*-
'''
===========
scrim.utils
===========
'''
from __future__ import absolute_import
import io
import os
from types import ModuleType
__all__ = [
    'this_path', 'relative_path', 'bin_path', 'copy_templates',
    'parse_setup', 'get_console_scripts', 'init_attr'
]

this_path = os.path.dirname(__file__)
NEWLINE_MAP = {
    '.ps1': '\r\n',
    '.bat': '\r\n',
    '.sh': '\n',
    '.csh': '\n',
    '.fish': '\n',
    '.zsh': '\n',
}


def relative_path(*args):
    '''os.path.join relative to this package'''

    return os.path.join(this_path, *args)


def bin_path(*args):
    '''os.path.join relative to this packages bin path'''

    return os.path.join(this_path, 'bin', *args)


def init_attr(value=None, default=None):
    '''Returns default if value is None'''
    if value is None:
        return default
    return value


def copy_templates(entry_point, py_entry_point, auto_write, output_dir):
    '''Copy formatted templates from scrim/bin to output directory

    Attributes:
        entry_point: Name of shell script
        py_entry_point: Name of python console script
        auto_write: Sets SCRIM_AUTO_WRITE to True
        output_dir: Guess
    '''

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    scripts = []
    for f in os.listdir(bin_path()):
        ext = os.path.splitext(f)[-1]
        newline = NEWLINE_MAP.get(ext, '\n')
        template = bin_path(f)
        destination = output_dir + '/' + entry_point + ext
        scripts.append(destination)

        with io.open(template, 'r') as f:
            code = f.read()
        code = code.replace('{{entry_point}}', entry_point)
        code = code.replace('{{py_entry_point}}', py_entry_point)
        code = code.replace('{{auto_write}}', str(int(auto_write)))

        with io.open(destination, 'w', newline=newline) as f:
            f.write(code)

    return scripts


def parse_setup(filepath):
    '''Get the kwargs from the setup function in setup.py'''
    # TODO: Need to parse setup.cfg and merge with the data from below

    # Monkey patch setuptools.setup to capture keyword arguments
    setup_kwargs = {}

    def setup_interceptor(**kwargs):
        setup_kwargs.update(kwargs)

    import setuptools
    setuptools_setup = setuptools.setup
    setuptools.setup = setup_interceptor

    # Manually compile setup.py
    with open(filepath, 'r') as f:
        code = compile(f.read(), '', 'exec')
    setup = ModuleType('setup')
    exec(code, setup.__dict__)

    # Remove monkey patch
    setuptools.setup = setuptools_setup
    return setup_kwargs


def get_console_scripts(setup_data):
    '''Parse and return a list of console_scripts from setup_data'''

    # TODO: support ini format of entry_points
    # TODO: support setup.cfg entry_points as available in pbr

    if 'entry_points' not in setup_data:
        return []

    console_scripts = setup_data['entry_points'].get('console_scripts', [])
    return [script.split('=')[0].strip() for script in console_scripts]



# -*- coding: utf-8 -*-
'''
==============
scrim.commands
==============
Implements functionality available across multiple shell scripting languages.
'''
from __future__ import absolute_import
import abc
from collections import namedtuple
from fstrings import f
import ntpath
import posixpath

ABC = abc.ABCMeta('ABC', (object,), {})
Command = namedtuple('Command', 'name args kwargs')
RawCommand = namedtuple('RawCommand', 'command required_shell')


class CommandExecutor(object):
    '''Forward commands to the specified ShellCommands implementation.
    RawCommands are returned as-is if the shell matches the RawCommands
    required_shell. You shouldn't need to use this class directly.
    '''

    def __call__(self, command, shell):

        if isinstance(command, Command):
            method = getattr(SHELL_COMMANDS[shell], command.name)
            return method(*command.args, **command.kwargs)

        elif isinstance(command, RawCommand):
            if shell == command.required_shell:
                return command.command

        else:
            raise TypeError(f(
                'command must be Command or RawCommand not {}', type(command)
            ))


class ShellCommands(ABC):
    '''Defines the interface for all ShellCommand implementations. These are
    the common commands we want to define for all shells.

    Sometimes commands are available across multiple shells like: cd. For
    clarity we define these commands for each implementation anyways instead of
    using inheritance.

    We only cover commands with functonality present in all the shells we
    support. Special cases can be handled using :meth:`Scrim.raw`.

    When choosing a method name for a command, generally use the most well
    known and widely used command. This will frequently come from bash. For
    example given the choice between `cat` in bash, `type` in batch, and
    `Get-Content` in powershell, we choose `cat` as the method name.

    In special cases we define methods that cover similar functionality for
    multiple shells like: `set_env` and `unset_env`. No shell has `set_env`
    or `unset_env` commands, but, powershell does have special syntax for
    setting and unsetting environment variables and bash has `export`.
    Therefore, we define `set_env` and `unset_env` for all implementations,
    even when batch has no direct analogy. This provides users with a
    memorable set of commands that cover all shells.

    Attributes:
        shell: Should match the SCRIM_SHELL value set in one of the scrim
            scripts. For example:

        - scrim.bat defines SCRIM_SHELL as cmd.exe
        - scrim.ps1 defines SCRIM_SHELL as powershell.exe
        - scrim.sh defines SCRIM_SHELL as bash
    '''

    @abc.abstractproperty
    def shell(self):
        raise NotImplementedError

    @abc.abstractmethod
    def execute(self, expression):
        raise NotImplementedError

    @abc.abstractmethod
    def echo(self, message):
        raise NotImplementedError

    @abc.abstractmethod
    def set(self, var, value):
        raise NotImplementedError

    @abc.abstractmethod
    def unset(self, var):
        raise NotImplementedError

    @abc.abstractmethod
    def set_env(self, var, value):
        raise NotImplementedError

    @abc.abstractmethod
    def unset_env(self, var):
        raise NotImplementedError

    @abc.abstractmethod
    def cd(self, path):
        raise NotImplementedError

    @abc.abstractmethod
    def pushd(self, path):
        raise NotImplementedError

    @abc.abstractmethod
    def popd(self):
        raise NotImplementedError

    @abc.abstractmethod
    def cat(self):
        raise NotImplementedError


class BatchCommands(ShellCommands):

    shell = 'cmd.exe'

    def execute(self, expression):
        return f('call {expression}')

    def echo(self, message):
        return f('echo {message}')

    def set(self, var, value):
        return f('set "{var}={value}"')

    def unset(self, var):
        return f('set "{var}="')

    set_env = set
    unset_env = unset

    def cd(self, path):
        path = ntpath.normpath(path)
        return f('cd {path}')

    def pushd(self, path):
        path = ntpath.normpath(path)
        return f('pushd {path}')

    def popd(self):
        return 'popd'

    def cat(self, path):
        path = ntpath.normpath(path)
        return f('type {path}')


class PowershellCommands(ShellCommands):

    shell = 'powershell.exe'

    def execute(self, expression):
        return f('Invoke-Expression {expression}')

    def echo(self, message):
        return f('Write-Host {message}')

    def set(self, var, value):
        return f('${var}={value}')

    def unset(self, var, value):
        return f('Remove-Variable {var}')

    def set_env(self, var, value):
        return f('$env:{var}={value}')

    def unset_env(self, var, value):
        return f('Remove-Item Env:{var}')

    def cd(self, path):
        path = ntpath.normpath(path)
        return f('cd {path}')

    def pushd(self, path):
        path = ntpath.normpath(path)
        return f('Push-Location -Path "{path}"')

    def popd(self):
        return 'Pop-Location'

    def cat(self, path):
        path = ntpath.normpath(path)
        return f('Get-Content {path}')


class BashCommands(ShellCommands):

    shell = 'bash'

    def execute(self, expression):
        return f('$({expression})')

    def echo(self, message):
        return f('echo {message}')

    def set(self, var, value):
        return f('{var}={value}')

    def unset(self, var):
        return f('unset {var}')

    def set_env(self, var, value):
        return f('export {var}={value}')

    def unset_env(self, var):
        return f('unset {var}')

    def cd(self, path):
        path = posixpath.normpath(path)
        return f('cd {path}')

    def pushd(self, path):
        path = posixpath.normpath(path)
        return f('pushd {path}')

    def popd(self):
        return 'popd'

    def cat(self, path):
        path = posixpath.normpath(path)
        return f('cat {path}')


SHELL_COMMANDS = dict((c.shell, c()) for c in ShellCommands.__subclasses__())



# -*- coding: utf-8 -*-
'''
=============
scrim.globals
=============
Defines variables passed into the python script via Environment Variables by
scrim scripts. If SCRIM_SHELL is None, then the python script was not executed
by a scrim script.

    SHELLS (list): list of available shells
    SCRIM_SHELL (str): Parent shell, one of the above SHELLS
    SCRIM_PATH (str): Path to output shell script
    SCRIM_AUTO_WRITE (bool): Write to SCRIM_PATH when python exits?
    SCRIM_SCRIPT (str): Path to the scrim script that invoked python
    SCRIM_DEBUG (bool): Is scrim script running in debug mode?
'''
from __future__ import absolute_import
import os
__all__ = [
    'SHELLS', 'SCRIM_SHELL', 'SCRIM_PATH', 'SCRIM_AUTO_WRITE',
    'SCRIM_SCRIPT', 'SCRIM_DEBUG'
]

SHELLS = [
    'powershell.exe',
    'cmd.exe',
    'bash'
]
SCRIM_SHELL = os.environ.get('SCRIM_SHELL', None)
SCRIM_PATH = os.environ.get('SCRIM_PATH', None)
SCRIM_AUTO_WRITE = bool(os.environ.get('SCRIM_AUTO_WRITE', False))
SCRIM_SCRIPT = os.environ.get('SCRIM_SCRIPT', None)
SCRIM_DEBUG = bool(os.environ.get('SCRIM_DEBUG', False))



# -*- coding: utf-8 -*-
'''
=========
scrim.api
=========
'''
from __future__ import absolute_import
import os
import atexit
from fstrings import f
from scrim.globals import (
    SHELLS,
    SCRIM_AUTO_WRITE,
    SCRIM_PATH,
    SCRIM_SHELL,
    SCRIM_SCRIPT
)
from scrim.commands import CommandExecutor, Command, RawCommand
from scrim.utils import init_attr
__all__ = ['Scrim', 'get_scrim']


try:
    basestring
except NameError:
    basestring = (str, bytes)


class Scrim(object):
    '''
    Arguments:
        shell: Name of the shell process. Defaulst to :envvar:`SCRIM_SHELL`
        path: Path to temporary scrim file. Defaults to :envvar:`SCRIM_PATH`
        auto_write: Write scrim file on program exit. Defaults to
                    :envvar:`SCRIM_AUTO_WRITE`
        script: Path to scrim script. Defaults to :envvar:`SCRIM_AUTO_WRITE`

    Usage:
        >>> scrim = Scrim()
        >>> scrim.echo('Hello World!')
        >>> scrim.to_cmd()
        'echo Hello World!'
    '''

    def __init__(self, path=None, auto_write=None, shell=None, script=None):
        self.shell = init_attr(shell, SCRIM_SHELL)
        self.path = init_attr(path, SCRIM_PATH)
        self.auto_write = init_attr(auto_write, SCRIM_AUTO_WRITE)
        self.script = init_attr(script, SCRIM_SCRIPT)
        self.command_executor = CommandExecutor()
        self.commands = []
        atexit.register(self.on_exit)

    def __repr__(self):
        args = (
            repr(self.path),
            repr(self.auto_write),
            repr(self.shell),
            repr(self.script)
        )
        return f("<{}>({}, {}, {}, {})", self.__class__.__name__, *args)

    def _add(self, name, *args, **kwargs):
        '''Appends a command to the scrims list of commands. You should not
        need to use this.'''

        self.commands.append(Command(name, args, kwargs))

    def execute(self, expression):
        '''Execute the specified expression or script.

        Arguments:
            expression (str): An expression or script file to execute
        '''

        self._add('execute', expression)

    def echo(self, message):
        '''Write to stdout.

        Arguments:
            message (str): Message to write
        '''

        self._add('echo', message)

    def set(self, var, value):
        '''Set a variable.

        Arguments:
            var (str): variable name
            value (Any): value to write
        '''

        self._add('set', var, value)

    def unset(self, var):
        '''Unset a variable.

        Arguments:
            var (str): variable name
        '''

        self._add('unset', var)

    def set_env(self, var, value):
        '''Set an environment variable. Depending on shell this is identical
        to :meth:`set`

        Arguments:
            var (str): variable name
            value (Any): value to write
        '''

        self._add('set_env', var, value)

    def unset_env(self, var):
        '''Unset an environment variable. Depending on shell this is identical
        to :meth:`unset`

        Arguments:
            var (str): variable name
        '''

        self._add('unset_env', var)

    def cd(self, path):
        '''Change directory.

        Arguments:
            path (str): Path to directory
        '''
        self._add('cd', path)

    def pushd(self, path):
        '''Push a directory onto the stack and navigate to it.

        Arguments:
            path (str): Path to directory
        '''
        self._add('pushd', path)

    def popd(self):
        '''Pop a directory from the stack and navigate to previous directory
        on stack
        '''

        self._add('popd')

    def raw(self, command, required_shell):
        '''Append a raw command to the scrim. This text will be appeneded to
        the output of :meth:`scrim.to_string` verbatim. Use required_shell to
        specify the shell which this raw command is applicable to.

        Examples:
            >>> scrim = Scrim()
            >>> scrim.raw('cat text.txt', 'bash')
            >>> scrim.raw('type text.txt', 'cmd.exe')
            >>> scrim.raw('Get-Content text.txt', 'powershell.exe')
            >>> scrim.to_bash()
            'cat text.txt'
            >>> scrim.to_cmd()
            'type text.txt'
            >>> scrim.to_powershell()
            'Get-Content text.txt'
        '''

        if not isinstance(command, basestring):
            raise TypeError(f('{command} must be a string'))

        if required_shell not in SHELLS:
            raise ValueError(f('{required_shell} must be one of {SHELLS}'))

        self.commands.append(RawCommand(command, required_shell))

    def to_string(self, shell=None):
        '''Use the command executor to retrieve the text of the scrim script
        compatible with the provided shell. If no shell is provided, use
        :attr:`scrim.shell`.

        Attributes:
            shell (str): Which shell should we return a script for
                cmd.exe, powershell.exe, bash...

        Returns:
            str
        '''

        shell = shell or self.shell
        lines = []
        for c in self.commands:
            text = self.command_executor(c, shell)
            if text is not None:
                lines.append(text)

        return '\n'.join(lines)

    def to_cmd(self):
        '''scrim.to_powershell() == scrim.to_string('powershell')

        Returns:
            str
        '''

        return self.to_string('cmd.exe')

    def to_powershell(self):
        '''scrim.to_powershell() == scrim.to_string('powershell')

        Returns:
            str
        '''

        return self.to_string('powershell.exe')

    def to_bash(self):
        '''scrim.to_bash() == scrim.to_string('bash')

        Returns:
            str
        '''

        return self.to_string('bash')

    def write(self):
        '''Write this Scrims commands to its path'''

        if self.path is None:
            raise Exception('Scrim.path is None')

        dirname = os.path.dirname(os.path.abspath(self.path))
        if not os.path.exists(dirname):
            try:
                os.makedirs(dirname)
            except:
                raise OSError('Failed to create root for scrim output.')

        with open(self.path, 'w') as f:
            f.write(self.to_string())

    def on_exit(self):
        '''atexit callback. If :attr:`Scrim.auto_write` is True write the
        scrim to :attr:`Scrim.path` as :attr:`Scrim.shell`'''

        if not all([self.auto_write, self.commands, self.script, self.path]):
            return

        self.write()


def get_scrim(path=None, auto_write=None, shell=None, script=None, cache={}):
    '''Get a :class:`Scrim` instance. Each instance is cached so if you call
    get_scrim again with the same arguments you get the same instance.

    See also:
        :class:`Scrim`
    '''

    args = (path, auto_write, shell, script)
    if args not in cache:
        cache[args] = Scrim(*args)
    return cache[args]



# -*- coding: utf-8 -*-
from __future__ import absolute_import

__title__ = 'scrim'
__author__ = 'Dan Bradham'
__email__ = 'danielbradham@gmail.com'
__license__ = 'MIT'
__url__ = 'https://github.com/danbradham/scrim.git'
__description__ = 'Opaque Python CLI Wrapper'
__version__ = '0.1.0'

from scrim.api import *
from scrim.globals import *
from scrim.utils import *
